/** @param {NS} ns */
export async function main(ns) {
	ns.clearLog()
	ns.tail()
	ns.disableLog('ALL')

	/** CCONFIG **/
	/* Maximum number of members in a gang */
	const MAX_MEMBER_COUNT = 12
	/* Percent of training to the next ascension a member must do after ascending */
	const MIN_ASCENSION_PERCENT = 0.95
	/* Minimum power level a member must reach before transitioning to gang war/money making */
	const MIN_POWER_LEVEL = 15
	/* Minimum win chance required to start a gang war */
	const GANG_WAR_START_TRESH = 0.6
	/* Minimum win chance before pulling out of a gang war */
	const GANG_WAR_STOP_TRESH = 0.55
	/* Target win chance */
	const GANG_WAR_TARGET = 0.95

	/** CONSTANTS **/
	const MAKE_MONEY = "Human Trafficking"
	const VIGILANTE_JUSTICE = "Vigilante Justice"
	const MUG_PEOPLE = "Mug People"
	const TERRORISM = "Terrorism"
	const TRAIN_COMBAT = "Train Combat"
	const TERRITORY_WARFARE = "Territory Warfare"

	const PHASE_CATCHUP = "catchup"
	const PHASE_RECRUIT = "recruit"
	const PHASE_TRAIN = "train"
	const PHASE_WAR = "gang war"
	const PHASE_MONEY = "make money"

	const memberNamePool = [
		"CJ",
		"Big Smoke",
		"Sweet",
		"Ryder",
		"OG Loc",
		"Big Bear",
		"Emmet",
		"Big Devil",
		"Little Devil",
		"Tony",
		"LB",
		"Madd Dogg",
		"Little Weasel",
	]
	const equipments = [
		"Baseball Bat",
		"Katana",
		"Glock 18C",
		"P90C", "Steyr AUG",
		"AK-47",
		"M15A10 Assault Rifle",
		"AWM Sniper Rifle",
		"Bulletproof Vest",
		"Full Body Armor",
		"Liquid Body Armor",
		"Graphene Plating Armor",
		"Ford Flex V20",
		"ATX1070 Superbike",
		"Mercedes-Benz S9001",
		"White Ferrari",
		"Bionic Arms",
		"Bionic Legs",
		"Bionic Spine",
		"BrachiBlades",
		"Nanofiber Weave",
		"Synthetic Heart",
		"Synfibril Muscle",
		"Graphene Bone Lacings",

		// Hacking stuff, low priority so they go to the end of the list
		"BitWire",
		"Neuralstimulator",
		"DataJack",
		"NUKE Rootkit",
		"Soulstealer Rootkit",
		"Demon Rootkit",
		"Hmap Node",
		"Jack the Ripper",
	]

	function getMinAscMult(memberName) {
		const memberInfo = ns.gang.getMemberInformation(memberName)
		return Math.min(
			memberInfo.str_asc_mult,
			memberInfo.def_asc_mult,
			memberInfo.dex_asc_mult,
			// memberInfo.agi_asc_mult,
		)
	}


	function getMinAscensionResult(memberName) {
		const memberAscensionResult = ns.gang.getAscensionResult(memberName)
		return memberAscensionResult === undefined ? undefined : Math.min(
			memberAscensionResult.str,
			memberAscensionResult.def,
			memberAscensionResult.dex,
			// memberAscensionResult.agi
		)
	}

	function estimateMemberPower(memberName) {
		return getMinAscMult(memberName) / 1.6
	}

	function getMinSkillLevel(memberName) {
		const memberInfo = ns.gang.getMemberInformation(memberName)
		return Math.min(
			memberInfo.str,
			memberInfo.def,
			memberInfo.dex,
			memberInfo.agi,
		)
	}

	// Credit: Mysteyes. https://discord.com/channels/415207508303544321/415207923506216971/940379724214075442
	function getAscendTreshold(memberName) {
		const mult = getMinAscMult(memberName)
		if (mult < 1.632) return 1.6326;
		if (mult < 2.336) return 1.4315;
		if (mult < 2.999) return 1.284;
		if (mult < 3.363) return 1.2125;
		if (mult < 4.253) return 1.1698;
		if (mult < 4.860) return 1.1428;
		if (mult < 5.455) return 1.1225;
		if (mult < 5.977) return 1.0957;
		if (mult < 6.496) return 1.0869;
		if (mult < 7.008) return 1.0789;
		if (mult < 7.519) return 1.073;
		if (mult < 8.025) return 1.0673;
		if (mult < 8.513) return 1.0631;
		return 1.0591;
	}

	function isRecruitmentReady(power, memberCount) {
		// the formula here is kinda arbitrary
		// increase the number of members in the gang linearly with the power of its members
		// ns.print(power * 1.5 + 3)
		return power * 1.5 + 3 > memberCount
	}

	function getGangWarWinChance() {
		const gangInfo = ns.gang.getGangInformation()
		const otherGangInfo = ns.gang.getOtherGangInformation()
		const otherGangNames = Object.keys(otherGangInfo).filter(x => x != gangInfo.faction && otherGangInfo[x]["territory"] > 0)
		const chancesToWin = otherGangNames.map(x => ns.gang.getChanceToWinClash(x))
		return Math.min(...chancesToWin)
	}

	while (true) {
		const gangInfo = ns.gang.getGangInformation()

		// check for and recruit new members if available
		while (ns.gang.canRecruitMember()) {
			const availableNames = memberNamePool.filter(x => memberNamePool.includes(x))
			const memberName = availableNames[Math.floor(Math.random() * availableNames.length)]
			ns.gang.recruitMember(memberName)
		}

		// gang war
		const gangWarWinChance = getGangWarWinChance()
		if (gangWarWinChance >= GANG_WAR_START_TRESH) {
			// go to war
			ns.gang.setTerritoryWarfare(true)
		}
		else if (gangWarWinChance <= GANG_WAR_STOP_TRESH) {
			// disengage
			ns.gang.setTerritoryWarfare(false)
		}

		// create an 2-dimensional array of member data. We use an array because we need it to be sorted
		// 1st dimension is for each member
		// index 0 of 2nd dimension is the name of the member
		// index 1 of 2nd dimension is the data
		let gangMembers = ns.gang.getMemberNames().map(x => [x, {
			"power": estimateMemberPower(x),
			"minSkillLevel": getMinSkillLevel(x),
			"ascensionPercent": getMinAscensionResult(x) / getAscendTreshold(x),
			"phase": undefined,
			"training": false,
			"vigilante": false,
		}])
		// sort by their power
		gangMembers.sort((a, b) => b[1]["power"] - a[1]["power"])
		// the highest power level is the target new members needs to catchup to
		const targetPower = gangMembers[0][1]["power"] - 2
		// check if we are ready to recruit a new member
		const recruitmentReady = isRecruitmentReady(targetPower, gangMembers.length)

		for (const member of gangMembers) {
			const memberName = member[0]
			const memberData = member[1]
			const power = memberData["power"]
			const minSkillLevel = memberData["minSkillLevel"]
			const ascensionPercent = memberData["ascensionPercent"]

			// compute the phase of each members
			if (power < targetPower) {
				// bring up newcomers to the same level than established members
				memberData["phase"] = PHASE_CATCHUP
			}
			else if (gangMembers.length < MAX_MEMBER_COUNT) {
				// gain respect to recruit new member
				memberData["phase"] = PHASE_RECRUIT
			}
			else if (power < MIN_POWER_LEVEL) {
				// power up until useful enough
				memberData["phase"] = PHASE_TRAIN
			}
			else if (gangInfo.territory < 1 && gangWarWinChance < GANG_WAR_TARGET) {
				// prepare for gang war
				memberData["phase"] = PHASE_WAR
			}
			else {
				// if everything else is done, make money
				memberData["phase"] = PHASE_MONEY
			}

			// assign tasks
			// strategy is:
			// catchup -> recruit <-> train -> gang war <-> make money
			// reduce wanted level as needed
			if (memberData["phase"] == PHASE_CATCHUP) {
				ns.print("assignment: catchup")
				ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
				memberData["training"] = true
			}
			else if (isNaN(ascensionPercent) || ascensionPercent < MIN_ASCENSION_PERCENT) { // TODO: arbitrary constant to be ajusted
				ns.print("assignment: level up after ascension")
				// train a member back up after ascention
				ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
				memberData["training"] = true
			}
			else if (gangInfo.wantedPenalty < 0) {
				ns.print("assignment: reduce wanted level")
				// try to keep the wanted penalty at 0%
				ns.gang.setMemberTask(memberName, VIGILANTE_JUSTICE)
				memberData["vigilante"] = true
			}
			else if (memberData["phase"] == PHASE_RECRUIT && recruitmentReady) {
				ns.print("assignment: gain respect to recruit new members")
				if (minSkillLevel < 250) {
					ns.gang.setMemberTask(memberName, MUG_PEOPLE)
				}
				else {
					ns.gang.setMemberTask(memberName, TERRORISM)
				}
			}
			else if (memberData["phase"] == PHASE_RECRUIT || memberData["phase"] == PHASE_TRAIN) {
				ns.print("assignment: train")
				// train up to next ascention
				ns.gang.setMemberTask(memberName, TRAIN_COMBAT)
				memberData["training"] = true
			}
			else if (memberData["phase"] == PHASE_WAR) {
				ns.print("assignment: gang war")
				ns.gang.setMemberTask(memberName, TERRITORY_WARFARE)
			}
			else {
				ns.print("assignment: make money")
				ns.gang.setMemberTask(memberName, MAKE_MONEY)
			}

			// buy equipement
			if (memberData["phase"] == PHASE_WAR || memberData["phase"] == PHASE_MONEY) {
				for (const equipment of equipments) {
					ns.gang.purchaseEquipment(memberName, equipment)
				}
			}

			// ascend members who are ready
			if (ascensionPercent >= 1 &&
				!(memberData["phase"] == PHASE_RECRUIT && memberData["training"] == false)) { // avoid ascending a member that has built up respect to recruit a new member
				// ns.print("ascending " + memberName)
				ns.gang.ascendMember(memberName)
			}
		}
		ns.print(gangMembers)
		await ns.sleep(5000)
	}
}